import { Env } from '../../../pi/lang/env';
import { Tr } from '../../../pi_pt/db/mgr';
import { Session } from '../../../pi_pt/net/session';
import { ResultBin, ResultStr } from '../../utils/result.s';
import { AddUser, Auth, Login, Passwd, TestReq, TestResp } from './user.s';

import { BonBuffer } from '../../../pi/util/bon';
import { mqttPublish, QoS } from '../../../pi_pt/rust/pi_serv/js_net';
import { Bucket } from '../../utils/db';
import { WARE_NAME } from '../data/constant';
import { AuthEnum, UserAuth } from '../data/db/auth.s';
import { User } from '../data/db/user.s';
import { writeLog } from './log.r';
import { WriteLog } from './log.s';

declare var env: Env;

/**
 * 用户管理
 */

// #[rpc=rpcServer]
export const login = (user: Login): ResultStr => {
    const log = new WriteLog();
    const b = new BonBuffer();
    log.arg = b.write(user.user).getBuffer();
    log.user = getUserName();
    log.ip = env.get('ip');
    log.module = 'management/server/rpc/user.s:login:Login';
    log.action = 'login';
    writeLog(log);
    
    const sess: Session = env.get('session');

    env.dbMgr.write((tr: Tr) => {
        sess.set(tr, 'userName', user.user);
    });

    const r = new ResultStr();

    const userBkt = new Bucket('file', User._$info.name);
    const u = userBkt.get<string, [User]>(user.user)[0];

    if (u && u.pwd === user.pwd) {
        r.code = 1;
        r.msg = 'login successfully';
    } else {
        r.code = 2;
        r.msg = 'incorrect user name or password';
    }

    return r;
};

/**
 * 添加用户
 */
// #[rpc=rpcServer]
export const addUser = (user: AddUser): ResultStr => {
    const log = new WriteLog();
    const b = new BonBuffer();
    log.arg = b.write(user.user).getBuffer();
    log.user = getUserName();
    log.ip = env.get('ip');
    log.module = 'management/server/rpc/user.s:addUser:string';
    log.action = 'add user';
    writeLog(log);

    const r = new ResultStr();

    // if (!checkAuth(AuthEnum.USER)) {
    //     r.code = 3;
    //     r.msg = 'auth failed';

    //     return r;
    // }
    
    const userBkt = new Bucket(WARE_NAME, User._$info.name);
    const res = userBkt.put(user.user, user);
    const userAuthBkt = new Bucket(WARE_NAME, UserAuth._$info.name);
    const ua = new UserAuth();
    ua.userName = user.user;
    ua.authList = [];
    userAuthBkt.put(user.user, ua);
    const buf = new BonBuffer();
    ua.bonEncode(buf);
    const mqttServer = env.get('mqttServer');
    mqttPublish(mqttServer,true,QoS.AtMostOnce,'admin',buf.getBuffer());

    if (res) {
        r.code = 1;
        r.msg = 'add user success';
    } else {
        r.code = 2;
        r.msg = 'add user failed';
    }

    return r;
};

/**
 * 修改用户权限
 */
// #[rpc=rpcServer]
export const modifyAuth = (user: Auth): ResultStr => {
    const log = new WriteLog();
    const b = new BonBuffer();
    user.bonEncode(b);
    log.arg = b.getBuffer();
    log.user = getUserName();
    log.ip = env.get('ip');
    log.module = 'management/server/rpc/user.s:modifyAuth:Auth';
    log.action = 'modify user auth';
    writeLog(log);

    const r = new ResultStr();
    const authBkt = new Bucket('file', UserAuth._$info.name);

    if (!checkAuth(AuthEnum.USER)) {
        r.code = 3;
        r.msg = 'auth failed';

        return r;
    }

    const res = authBkt.put(user.user, user);
    const buf = new BonBuffer();
    user.bonEncode(buf);
    const mqttServer = env.get('mqttServer');
    mqttPublish(mqttServer,true,QoS.AtMostOnce,'auth',buf.getBuffer());

    if (res) {
        r.code = 1;
        r.msg = 'modifyAuth success';
    } else {
        r.code = 2;
        r.msg = 'modifyAuth failed';
    }

    return r;
};

/**
 * 删除用户
 */
// #[rpc=rpcServer]
export const delUser = (user: string): ResultStr => {
    const log = new WriteLog();
    const b = new BonBuffer();
    log.arg = b.writeUtf8(user).getBuffer();
    log.user = getUserName();
    log.ip = env.get('ip');
    log.module = 'management/server/rpc/user.s:delUser:string';
    log.action = 'delete user';
    writeLog(log);

    const r = new ResultStr();
    
    // if (!checkAuth(AuthEnum.USER) || user === 'admin') {
    //     r.code = 3;
    //     r.msg = 'auth failed';
        
    //     return r;
    // }
    const userBkt = new Bucket(WARE_NAME, User._$info.name);
    const res = userBkt.delete(user);
    const userAuthBkt = new Bucket(WARE_NAME, UserAuth._$info.name);
    userAuthBkt.delete(user);

    if (res) {
        r.code = 1;
        r.msg = 'delete user success';
    } else {
        r.code = 2;
        r.msg = 'delete user failed';
    }

    return r;
};

// #[rpc=rpcServer]
export const changePasswd = (passwd: Passwd): ResultStr => {
    const log = new WriteLog();
    log.user = getUserName();
    log.ip = env.get('ip');
    log.module = 'management/server/rpc/user.s:changePasswd:Passwd';
    log.action = 'change passwd';
    writeLog(log);

    const r = new ResultStr();
    const userBkt = new Bucket(WARE_NAME, User._$info.name);
    const user = getUserName();
    const u = userBkt.get<string, [User]>(user)[0];

    if (u && (u.pwd === passwd.old)) {
        const uu = new User;
        uu.user = u.user;
        uu.pwd = passwd.new;
        if (userBkt.put(user, uu)) {
            r.code = 1;
            r.msg = 'change passwd successfully';
        } else {
            r.code = 2;
            r.msg = 'change passwd failed';
        }
    } else {
        r.code = 2;
        r.msg = 'change passwd failed';
    }

    return r;
};

// #[rpc=rpcServer]
export const listAllUser = (): ResultBin => {
    const res = new ResultBin();
    const userAuthBkt = new Bucket(WARE_NAME, UserAuth._$info.name);
    const iter = userAuthBkt.iter(null);
    const users = [];
    while (true) {
        const elem = iter.next();
        if (!elem)break;
        users.push(elem[1]);
    }

    const bon = new BonBuffer();
    bon.writeArray(users, (el: UserAuth) => {
        const b = new BonBuffer();
        el.bonEncode(b);
        bon.writeBin(b.getBuffer());
    });

    res.code = 1;
    res.msg = bon.getBuffer();

    return res;
};

// #[rpc=rpcServer]
export const getUserAuth = (): UserAuth => {
    const userAuthBkt = new Bucket(WARE_NAME, UserAuth._$info.name);
    let user = getUserName();
    let auth = userAuthBkt.get<string, [UserAuth]>(user)[0]

    return auth;
}

/**
 * 验证用户权限
 */
export const checkAuth = (auth: number): boolean => {
    let userName;

    const sess: Session = env.get('session');
    env.dbMgr.read((tr: Tr) => {
        userName = sess.get(tr, 'userName');
    });

    if (!userName) return false;

    const authBkt = new Bucket('file', UserAuth._$info.name);
    const auth1 = authBkt.get<string, [UserAuth]>(userName)[0];
    if (auth1 && auth1.authList.indexOf(auth) > -1) {
        return true;
    }

    return false;
};

export const getUserName = () => {
    let userName;
    const sess: Session = env.get('session');
    env.dbMgr.read((tr: Tr) => {
        userName = sess.get(tr, 'userName');
    });
    return userName;
};

// ================== test rpcs =============== 

// #[rpc=rpcServer]
export const testRpc = (req: TestReq): TestResp => {
    const res = new TestResp;
    res.code = 1;
    res.msg = 'test rpc';

    return res;
};

// #[rpc=rpcServer]
export const testResultBin = (): ResultBin => {
    const res = new ResultBin;
    const bon = new BonBuffer(new Uint8Array([1,2,3]));
    res.code = 1;
    res.msg = bon.getBuffer();

    return res;
};