/// <reference path="./babylon.d.ts"/>
import {ResTab, Res, register, loadOK} from "../../util/res_mgr";
import {Geometry} from "./geometry";
import {OfflineProvider} from "./offline_provider";

export class ResLoader {
    resTab: ResTab;

    constructor(resTab: ResTab){
        this.resTab = resTab;
    }

    public loadTexture (url: BABYLON.Nullable<string>, noMipmap: boolean, sampling?: number): BABYLON.Nullable<BABYLON.InternalTexture> {
        let r = this.resTab.get(textureKey(url, noMipmap, sampling)) as TextureRes;
        if (r) {
            return r.link;
        }else {
            return null;
        }
    }

    public createTexture (texture: BABYLON.InternalTexture) {
        //该load 是一个同步过程
        this.resTab.load(textureKey(texture.url, !texture.generateMipMaps, texture.samplingMode), BABYLON_TEXTURE_TYPE, texture, null, null, () => {
            console.log("createTexture fail!, url:", texture.url, "noMipmap:", texture.generateMipMaps,"sampling:", texture.samplingMode);
        });
    }

    //调用此方法应该保证resTab中存在该texture
    public releaseTexture(texture: BABYLON.InternalTexture): boolean{
        if(!texture.url){
            return false;
        }
        let res = this.resTab.get(textureKey(texture.url, !texture.generateMipMaps, texture.samplingMode)); 
        if(!res){
            return false;
        }
        this.resTab.delete(res);
        return true;
    }

    public loadGeometry(id: string): Geometry{
        let r = this.resTab.get(geometryKey(id)) as GeometryRes;
        if (r) {
            return r.link;
        }else {
            return null;
        }
    }

    public createGeometry(geometry: Geometry){
        //该load 是一个同步过程
        this.resTab.load(geometryKey(geometry.id), BABYLON_GEOMETRY_TYPE, geometry, null, null, () => {
            console.log("createGeometry fail!, url:", geometry.id);
        });
    }

     //调用此方法应该保证resTab中存在该texture
     public releaseGeometry(id: string){
        let res = this.resTab.get(geometryKey(id)); 
        this.resTab.delete(res);
    }
}

/**
 * @description 纹理资源
 * @example
 */
export class TextureRes extends Res {
    link: BABYLON.InternalTexture;
	/**
	 * @description 创建
	 * @example
	 */
	public create(data: BABYLON.InternalTexture): void {
		this.link = data;
	}

	/**
	 * @description 销毁，需要子类重载
	 * @example
	 */
	public destroy(): void {
        let r = this.link;
		if (r) {
            (<any>r)._engine._releaseTexture(r);
            r._webGLTexture = null;
            r.previous = null;
            r.next = null;
		}
	}
}

/**
 * @description 纹理资源
 * @example
 */
export class GeometryRes extends Res {
    link: Geometry;
	/**
	 * @description 创建
	 * @example
	 */
	public create(data: Geometry): void {
		this.link = data;
	}

	/**
	 * @description 销毁，需要子类重载
	 * @example
	 */
	public destroy(): void {
        let r = this.link;
		if (r) {
            BABYLON.Geometry.prototype.dispose.call(r)
		}
	}
}

/**
 * 对Babylon的GLTFLoader进行扩展， 其目的是接管其VertexBuffer的加载, 对其进行缓冲！
 * 使用该扩展类， 你始终应该保证accessor引用的buffer是紧凑的数据（accessor没有bytesStride字段）， 并保证数据的offset可以整除数据类型的长度
 */
export class GLTF2Loader extends BABYLON.GLTF2.GLTFLoader{
    static resLoader: ResLoader;

    //重载_loadVertexBufferViewAsync方法， 使其返回值为ArrayBufferView
   _loadVertexBufferViewAsync(bufferView: BABYLON.GLTF2.Loader.IBufferView, _kind: string): Promise<ArrayBufferView> {
        if (bufferView._babylonBuffer) {
            return bufferView._babylonBuffer as any;
        }

        bufferView._babylonBuffer = this.loadBufferViewAsync(`#/bufferViews/${bufferView.index}`, bufferView).then((data) => {
            return data as any;
        });

        return bufferView._babylonBuffer as any;
    }

    //重载_loadVertexAccessorAsync方法， 使其不支持sparse， 并且偏移量只能是其类型长度的倍数
    _loadVertexAccessorAsync(context: string, accessor: BABYLON.GLTF2.Loader.IAccessor, kind: string): Promise<BABYLON.VertexBuffer> {
        if (accessor._babylonVertexBuffer) {
            return accessor._babylonVertexBuffer;
        }

        if (accessor.sparse || (accessor.byteOffset && accessor.byteOffset % BABYLON.VertexBuffer.GetTypeByteLength(accessor.componentType) !== 0)) {
            throw new Error("Accessor does not support s, and its cost should be integer multiple of its type.");
        }

        const bufferView = BABYLON.GLTF2.ArrayItem.Get(`${context}/bufferView`, this.gltf.bufferViews, accessor.bufferView);
        accessor._babylonVertexBuffer = this._loadVertexBufferViewAsync(bufferView, kind).then((babylonBuffer) => {
            const size = (<any>BABYLON.GLTF2.GLTFLoader)._GetNumComponents(context, accessor.type);
            return new BABYLON.VertexBuffer(this.babylonScene.getEngine(), babylonBuffer, kind, false, false, null,
                false, accessor.byteOffset, size, accessor.componentType, accessor.normalized, true);
        });

        return accessor._babylonVertexBuffer;
    }

    //重载_loadVertexDataAsync， 缓存geo
    _loadVertexDataAsync(context: string, primitive:  BABYLON.GLTF2.Loader.IMeshPrimitive, babylonMesh: BABYLON.Mesh): Promise<BABYLON.Geometry> {
        const accessor = BABYLON.GLTF2.ArrayItem.Get(`${context}/indices`, this.gltf.accessors, primitive.indices);
        const bufferView = BABYLON.GLTF2.ArrayItem.Get(`#/accessors/${accessor.index}/bufferView`, this.gltf.bufferViews, accessor.bufferView);
        const buffer = BABYLON.GLTF2.ArrayItem.Get(`#/bufferViews/${bufferView.index}/buffer`, this.gltf.buffers, bufferView.buffer);
        let url = (<any>this)._rootUrl + buffer.uri;
        let r = GLTF2Loader.resLoader.loadGeometry(url);
        if (r){
            return new Promise(() => {return r});
        }else {
            let the = this as any;
            const extensionPromise = the._extensionsLoadVertexDataAsync(context, primitive, babylonMesh);
            if (extensionPromise) {
                return extensionPromise;
            }

            const attributes = primitive.attributes;
            if (!attributes) {
                throw new Error(`${context}: Attributes are missing`);
            }

            const promises = new Array<Promise<any>>();

            const babylonGeometry = new Geometry(url, this.babylonScene);
            GLTF2Loader.resLoader.createGeometry(babylonGeometry);

            if (primitive.indices == undefined) {
                babylonMesh.isUnIndexed = true;
            }
            else {
                promises.push(the._loadIndicesAccessorAsync(`#/accessors/${accessor.index}`, accessor).then((data) => {
                    babylonGeometry.setIndices(data);
                }));
            }

            const loadAttribute = (attribute: string, kind: string, callback?: (accessor: BABYLON.GLTF2.Loader.IAccessor) => void) => {
                if (attributes[attribute] == undefined) {
                    return;
                }

                babylonMesh._delayInfo = babylonMesh._delayInfo || [];
                if (babylonMesh._delayInfo.indexOf(kind) === -1) {
                    babylonMesh._delayInfo.push(kind);
                }

                const accessor = BABYLON.GLTF2.ArrayItem.Get(`${context}/attributes/${attribute}`, this.gltf.accessors, attributes[attribute]);
                promises.push(this._loadVertexAccessorAsync(`#/accessors/${accessor.index}`, accessor, kind).then((babylonVertexBuffer) => {
                    babylonGeometry.setVerticesBuffer(babylonVertexBuffer, accessor.count);
                }));

                if (callback) {
                    callback(accessor);
                }
            };

            loadAttribute("POSITION", BABYLON.VertexBuffer.PositionKind);
            loadAttribute("NORMAL", BABYLON.VertexBuffer.NormalKind);
            loadAttribute("TANGENT", BABYLON.VertexBuffer.TangentKind);
            loadAttribute("TEXCOORD_0", BABYLON.VertexBuffer.UVKind);
            loadAttribute("TEXCOORD_1", BABYLON.VertexBuffer.UV2Kind);
            loadAttribute("JOINTS_0", BABYLON.VertexBuffer.MatricesIndicesKind);
            loadAttribute("WEIGHTS_0", BABYLON.VertexBuffer.MatricesWeightsKind);
            loadAttribute("COLOR_0", BABYLON.VertexBuffer.ColorKind, (accessor) => {
                if (accessor.type === BABYLON.GLTF2.AccessorType.VEC4) {
                    babylonMesh.hasVertexAlpha = true;
                }
            });

            return Promise.all(promises).then(() => {
                return babylonGeometry;
            }); 
        }
    }

    //重载_loadAnimationChannelAsync， 缓存keys
    _loadAnimationChannelAsync(context: string, animationContext: string, animation: BABYLON.GLTF2.Loader.IAnimation, channel: BABYLON.GLTF2.Loader.IAnimationChannel, babylonAnimationGroup: BABYLON.AnimationGroup): Promise<void> {
        const targetNode = BABYLON.GLTF2.ArrayItem.Get(`${context}/target/node`, this.gltf.nodes, channel.target.node);

        // Ignore animations that have no animation targets.
        if ((channel.target.path === BABYLON.GLTF2.AnimationChannelTargetPath.WEIGHTS && !targetNode._numMorphTargets) ||
            (channel.target.path !== BABYLON.GLTF2.AnimationChannelTargetPath.WEIGHTS && !targetNode._babylonMesh)) {
            return Promise.resolve();
        }

        // Ignore animations targeting TRS of skinned nodes.
        // See https://github.com/KhronosGroup/glTF/tree/master/specification/2.0#skins (second implementation note)
        if (targetNode.skin != undefined && channel.target.path !== BABYLON.GLTF2.AnimationChannelTargetPath.WEIGHTS) {
            return Promise.resolve();
        }

        const sampler = BABYLON.GLTF2.ArrayItem.Get(`${context}/sampler`, animation.samplers, channel.sampler);
        const inputAccessor = BABYLON.GLTF2.ArrayItem.Get(`${animationContext}/samplers/${channel.sampler}/input`, this.gltf.accessors, sampler.input);
        const bufferView = BABYLON.GLTF2.ArrayItem.Get(`#/bufferView`, this.gltf.bufferViews, inputAccessor.bufferView);
        const buffer = BABYLON.GLTF2.ArrayItem.Get(`#/bufferViews/${bufferView.index}/buffer`, this.gltf.buffers, bufferView.buffer);
        let url = (<any>this)._rootUrl + buffer.uri;

        let targetPath: string;
        let animationType: number;
        switch (channel.target.path) {
            case BABYLON.GLTF2.AnimationChannelTargetPath.TRANSLATION: {
                targetPath = "position";
                animationType = BABYLON.Animation.ANIMATIONTYPE_VECTOR3;
                break;
            }
            case BABYLON.GLTF2.AnimationChannelTargetPath.ROTATION: {
                targetPath = "rotationQuaternion";
                animationType = BABYLON.Animation.ANIMATIONTYPE_QUATERNION;
                break;
            }
            case BABYLON.GLTF2.AnimationChannelTargetPath.SCALE: {
                targetPath = "scaling";
                animationType = BABYLON.Animation.ANIMATIONTYPE_VECTOR3;
                break;
            }
            case BABYLON.GLTF2.AnimationChannelTargetPath.WEIGHTS: {
                targetPath = "influence";
                animationType = BABYLON.Animation.ANIMATIONTYPE_FLOAT;
                break;
            }
            default: {
                throw new Error(`${context}/target/path: Invalid value (${channel.target.path})`);
            }
        }

        const influenceAnim = (keys) => {
            for (let targetIndex = 0; targetIndex < targetNode._numMorphTargets!; targetIndex++) {
                const animationName = `${babylonAnimationGroup.name}_channel${babylonAnimationGroup.targetedAnimations.length}`;
                const babylonAnimation = new BABYLON.Animation(animationName, targetPath, 1, animationType);
                let iKeys = (<OfflineProvider>this.babylonScene.offlineProvider).loadAnimationKeys(url + "_" + 0);
                if(!iKeys){
                    iKeys = keys.map((key) => ({
                        frame: key.frame,
                        inTangent: key.inTangent ? key.inTangent[targetIndex] : undefined,
                        value: key.value[targetIndex],
                        outTangent: key.outTangent ? key.outTangent[targetIndex] : undefined
                    }));
                    (<OfflineProvider>this.babylonScene.offlineProvider).createAnimationKeys(url + "_" + 0, iKeys);
                }

                babylonAnimation.setKeys(iKeys);
                (<any>this)._forEachPrimitive(targetNode, (babylonMesh) => {
                    const morphTarget = babylonMesh.morphTargetManager!.getTarget(targetIndex);
                    const babylonAnimationClone = babylonAnimation.clone();
                    morphTarget.animations.push(babylonAnimationClone);
                    babylonAnimationGroup.addTargetedAnimation(babylonAnimationClone, morphTarget);
                });
            }
        }

        if (targetPath === "influence") {
            if(!(<OfflineProvider>this.babylonScene.offlineProvider).loadAnimationKeys(url + "_" + 0)){
                (<any>this)._loadAnimationKeysAsync(context, animationContext, animation, channel, url).then((keys) => {
                    influenceAnim(keys);
                });
            }else {
                influenceAnim(null);
            }
        } else {
            (<any>this)._loadAnimationKeysAsync(context, animationContext, animation, channel, url).then((keys) => {
                const animationName = `${babylonAnimationGroup.name}_channel${babylonAnimationGroup.targetedAnimations.length}`;
                const babylonAnimation = new BABYLON.Animation(animationName, targetPath, 1, animationType);
                babylonAnimation.setKeys(keys);
    
                if (targetNode._babylonBones) {
                    const babylonAnimationTargets = [targetNode._babylonMesh!, ...targetNode._babylonBones];
                    for (const babylonAnimationTarget of babylonAnimationTargets) {
                        babylonAnimationTarget.animations.push(babylonAnimation);
                    }
                    babylonAnimationGroup.addTargetedAnimation(babylonAnimation, babylonAnimationTargets);
                }
                else {
                    targetNode._babylonMesh!.animations.push(babylonAnimation);
                    babylonAnimationGroup.addTargetedAnimation(babylonAnimation, targetNode._babylonMesh);
                }
            });
        }
    }

     //增加_loadAnimationKeysAsync方法， 缓存keys
     _loadAnimationKeysAsync(context: string, animationContext: string, animation: BABYLON.GLTF2.Loader.IAnimation, channel: BABYLON.GLTF2.Loader.IAnimationChannel, key: string): Promise<void> {
        let r = (<OfflineProvider>this.babylonScene.offlineProvider).loadAnimationKeys(key);
        if(r){
            return new Promise(() => {return r});
        }

        const targetNode = BABYLON.GLTF2.ArrayItem.Get(`${context}/target/node`, this.gltf.nodes, channel.target.node);
        let targetPath: string;

        const sampler = BABYLON.GLTF2.ArrayItem.Get(`${context}/sampler`, animation.samplers, channel.sampler);
        return (<any>this)._loadAnimationSamplerAsync(`${animationContext}/samplers/${channel.sampler}`, sampler).then((data) => {
            let outputBufferOffset = 0;
            let getNextOutputValue: () => BABYLON.Vector3 | BABYLON.Quaternion | Array<number>;
            switch (targetPath) {
                case "position": {
                    getNextOutputValue = () => {
                        const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 3;
                        return value;
                    };
                    break;
                }
                case "rotationQuaternion": {
                    getNextOutputValue = () => {
                        const value = BABYLON.Quaternion.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 4;
                        return value;
                    };
                    break;
                }
                case "scaling": {
                    getNextOutputValue = () => {
                        const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                        outputBufferOffset += 3;
                        return value;
                    };
                    break;
                }
                case "influence": {
                    getNextOutputValue = () => {
                        const value = new Array<number>(targetNode._numMorphTargets!);
                        for (let i = 0; i < targetNode._numMorphTargets!; i++) {
                            value[i] = data.output[outputBufferOffset++];
                        }
                        return value;
                    };
                    break;
                }
            }

            let getNextKey: (frameIndex: number) => BABYLON.IAnimationKey;
            switch (data.interpolation) {
                case BABYLON.GLTF2.AnimationSamplerInterpolation.STEP: {
                    getNextKey = (frameIndex) => ({
                        frame: data.input[frameIndex],
                        value: getNextOutputValue(),
                        interpolation:  BABYLON.AnimationKeyInterpolation.STEP
                    });
                    break;
                }
                case BABYLON.GLTF2.AnimationSamplerInterpolation.LINEAR: {
                    getNextKey = (frameIndex) => ({
                        frame: data.input[frameIndex],
                        value: getNextOutputValue()
                    });
                    break;
                }
                case BABYLON.GLTF2.AnimationSamplerInterpolation.CUBICSPLINE: {
                    getNextKey = (frameIndex) => ({
                        frame: data.input[frameIndex],
                        inTangent: getNextOutputValue(),
                        value: getNextOutputValue(),
                        outTangent: getNextOutputValue()
                    });
                    break;
                }
            }

            const keys = new Array(data.input.length);
            for (let frameIndex = 0; frameIndex < data.input.length; frameIndex++) {
                keys[frameIndex] = getNextKey!(frameIndex);
            }

            (<OfflineProvider>this.babylonScene.offlineProvider).createAnimationKeys(key, keys); //缓存关键帧
            return keys;
        });
    }
}

const textureKey = (url: BABYLON.Nullable<string>, noMipmap: boolean, sampling?: number): string => {
    return BABYLON_TEXTURE_TYPE + ":" + url + "_" + noMipmap + "_" + (sampling?true:false);
}

const geometryKey = (id: string): string => {
    return BABYLON_GEOMETRY_TYPE + ":" + id;
}

const createTextureRes = (name: string, type: string, texture: BABYLON.InternalTexture, _: any) => {
	return loadOK(name, type, null, TextureRes, texture);
};

const createGeometryRes = (name: string, type: string, geometry: Geometry, _: any) => {
	return loadOK(name, type, null, GeometryRes, geometry);
};

const BABYLON_TEXTURE_TYPE = "babylon_texture";
const BABYLON_GEOMETRY_TYPE = "babylon_geometry";

register(BABYLON_TEXTURE_TYPE, (name: string, type: string, texture: BABYLON.InternalTexture, _: any) => {
	createTextureRes(name, type, texture, _);
});

register(BABYLON_GEOMETRY_TYPE, (name: string, type: string, geometry: Geometry, _: any) => {
	createGeometryRes(name, type, geometry, _);
});