// 模块描述
/*
负责显示逻辑，是数据和原始dom间的桥梁
组件支持嵌套，并且tpl中的自定义元素支持相对路径。 
组件名的规则：可以使用英文小写字母加'_'和''。 '-'表示路径分隔，'$'只能在最后，1个'$'表示本目录开始查找，N个'$'表示上溯N-1个父目录开始查找。如果没有'$'表示从根目录下开始查找
举例：
<role_show$ style=""></role_show$>表示本目录下的role_show组件，
<role_show$$ style=""> </role_show$$>表示父目录下的role_show组件，
<role_show-zb_show$$ style=""></role_show-zb_show$$>表示父目录下role_show目录下的zb_show组件
<app-base-btn style=""></app-base-btn>表示根目录开始，app/base目录下的btn组件
*/
// tslint:disable-next-line:no-reference
/// <reference path="../render3d/babylon/babylon.d.ts"/>

// ============================== 导入
import { Json } from '../lang/type';
import { GUICreator } from '../render3d/babylon/gui_creator';
import { Widget } from './widget';

// ============================== 导出

/**
 * @description 组件
 * @example
 * 组件，包含样式和模板的js类,
 * 注意区分 widget实例和widget节点
 * widget节点的link属性指向了widget实例
 */
export class WidgetGUI extends Widget {
    // 必须要赋初值，不然new出来的实例里面是没有这些属性的

    public control: BABYLON.GUI.Container; //   GUI 节点

    /**
     * 创建 GUI 节点
     */
    public createGUI: () => BABYLON.GUI.Container;

    // gui 节点更新方法
    public guiUpdate: () => void;

    /**
     * 创建后调用，一般在渲染循环外调用
     */
    public create(): void {

        // 屏蔽 tpl 的使用
        this.tpl = undefined;

        this.forelet && this.forelet.addWidget(this);

    }

    /**
     * 设置属性，默认外部传入的props是完整的props，重载可改变行为
     * @param props 新数据
     * @param oldProps 老数据
     */
    public setProps(props: Json, oldProps?: Json): void {

        this.props = props;

        if (props !== undefined && props.GUI) {

            this._createGUI();

            this.control && GUICreator.getUIRoot().addControl(this.control);
            
        }
        
        this.props.GUI = undefined;
    }
    
    /**
     * 第一次计算后调用，此时创建了真实的dom，但并没有加入到dom树上，一般在渲染循环外调用
     */
    public firstPaint(): void {
        // this.forelet && this.forelet.eventWidget(this, 'firstPaint');
    }
    /**
     * 销毁时调用，一般在渲染循环外调用
     */
    public destroy(): boolean {
        // 销毁GUI节点
        // if (this.control !== undefined) {
        //     this.control.dispose();
        //     this.control = undefined;
        // }
        const control = this.control;
        this.control = undefined;
        this.children.length = 0;
        GUICreator.disposeWithTree(control);

        this.forelet && this.forelet.removeWidget(this);

        if (!this.tpl) {
            return false;
        }
        
        this.tpl = undefined;
        if (this.resTab) {
            this.resTab.timeout = this.resTimeout;
            this.resTab.release();
        }

        return true;
    }
    /**
     * 绘制方法，
     * @param reset 表示新旧数据差异很大，不做差异计算，直接生成dom
     */
    public paint(reset?: boolean): void {

        // paintWidget(this, reset);
        
        this._guiUpdate();
    }

    private _createGUI() {
        this.control === undefined && this.createGUI && this.createGUI();
    }
    private _guiUpdate() {
        this.control && this.guiUpdate && this.guiUpdate();
    }

}
