/**
 * 数据库初始化
 * 根据DbCfg配置，创建数据库
 * DbCfg:{
 *      name: string, //为数据库取一个名称
 *      dbType: DbType //数据库的类型， 支持文件数据库和内存数据库， 如果是文件数据库需要一个path(文件库的路径)
 * }
 *
 * 根据struct上的注解，初始化数据库表
 * 在struct添加"db"注解， 表示该类型的实例会存入数据库， 会自动初始化数据的表
 * 注解规则 :  #[db=数据库名称]
 *
 * 例如： 在.c.ts文件中设置了一个配置：
 *         new DbCfg("db1", new DbType(DbType.Memory)) //该配置表明需要注册一个内存数据库， 数据库名称为“db1”
 *
 * 在.rs中定义结构体：
 *         //在结构体中添加“db”注解， 表示数据库“db1”中需要创建Player的表
 *         #[db=db1]
 *         struct Player{}
 *
 * 该模块将配置与注解结合起来， 使得数据库及数据库表的创建自动化
 */

import { DbCfg} from "./server_cfg.s";
import { restore, /*registerFileDb*/} from "../rust/pi_serv/js_db";
import { Mgr} from "../../pi/db/mgr";
import { DB as MemeryDb} from "../rust/pi_db/memery_db";
import { DB as FileDb} from "../rust/pi_store/lmdb_file";
import { FileMemDB } from "../rust/pi_store/file_mem_db";
import { Struct } from "../../pi/struct/struct_mgr";
import { StructInfo, EnumInfo , TabMeta, EnumType, Type, FieldInfo} from "../../pi/struct/sinfo";
import { cfgMgr } from "../../pi/util/cfg";
import { DEFAULT_WARE, DEFAULT_FILE_WARE, FILE_MEM_WARE } from "../constant";
import { SESSION_WARE } from "../constant";
import {Env} from "../../pi/lang/env";
import {} from "./init";
declare var env: Env;


declare var pi_modules;
const initDb = () => {
    let dbMgr = env.dbMgr;
    // 创建session库
    dbMgr.register(SESSION_WARE, MemeryDb.new());
    // dbMgr.register(DEFAULT_FILE_WARE, MemeryDb.new());
    // registerMemeryDb(dbMgr, SESSION_WARE, MemeryDb.new())
    //默认创建文件数据库
    // dbMgr.register(DEFAULT_FILE_WARE, FileDb.new(DEFAULT_FILE_WARE, 100 * 1024 * 1024));
    dbMgr.register(DEFAULT_FILE_WARE, FileMemDB.new(DEFAULT_FILE_WARE, 1000 * 1024 * 1024));
    // registerMemeryDb(dbMgr, DEFAULT_FILE_WARE , MemeryDb.new());
    // Lmdb
    // registerFileDb(dbMgr, LMDB_WARE, Lmdb.new("lmdb", 1024*1024*100));
    //取到数据库服务配置
    let cfgs = cfgMgr.getPrimary(DbCfg._$info.name, "name");
    if(cfgs){
        //遍历配置，注册对应的数据库
        cfgs.forEach((v, k) => {
            if(k === DEFAULT_FILE_WARE || k === DEFAULT_WARE){
                throw new Error(`${k} is a default db, is not create`);
            }
            dbMgr.register(k as string, MemeryDb.new());
            // registerMemeryDb(dbMgr, k as string , MemeryDb.new());
        });
    }
    //遍历所有的Struct， 根据注解创建数据库表
    dbMgr.write((tr) => {
        for(var id in pi_modules){
            if(!pi_modules.hasOwnProperty(id) || !pi_modules[id].exports){
                continue;
            }
            for(var kk in pi_modules[id].exports){
                var c = pi_modules[id].exports[kk];
                if(!Struct.isPrototypeOf(c) || !c._$info){
                    continue;
                }

                let notes = c._$info.notes || new Map();
                let dbName = notes.get("db") || DEFAULT_WARE; //如果没有“db注解”， 使用默认的数据库

                let key_type, value_type;
                //key 元信息根据主键确定

                let primary = notes.get("primary");
                if (primary){
                    key_type = getFieldType(primary, c._$info.fields); //如果存在主键， 数据库表的key类型为主键的类型
                }else{
                    key_type = new EnumType(Type.Usize);//否则为索引的类型
                }

                //value 元信息根据_$info确定
                if(c._$info instanceof StructInfo){
                    value_type = new EnumType(Type.Struct, c._$info);
                }else if(c._$info instanceof EnumInfo){
                    value_type = new EnumType(Type.Enum, c._$info);
                }
                //初始化是根据Struct创建数据库表
                tr.alter(dbName, (<StructInfo>c._$info).name, new TabMeta(key_type, value_type));
            }

        }
    }, 500000);
    //写配置数据
    writeCfg(dbMgr);
    //数据库恢复， 临时方案
    // restore_temp(cfgs as any, dbMgr);
}

const getFieldType = (name: string, fields: FieldInfo[]): EnumType => {
    for(let i = 0; i < fields.length; i++){
        if(fields[i].name === name){
            return fields[i].ftype;
        }
    }
    throw new Error("field is not exist, name: " + name);
}

/**
 * 恢复数据库
 * 由于底层的文件数据库不完备， 暂时使用内存数据代替文件数据库使用， 每次数据库的改变将会备份到文件中， 当重新启动平台是需要从文件中恢复到内存数据库
 * 此方法就是启动过程中数据库的恢复
 * */
const restore_temp = (cfgs: Map<string, DbCfg>, dbMgr: Mgr) => {
    //临时方案， 恢复数据库
    for(var id in pi_modules){
        if(!pi_modules.hasOwnProperty(id) || !pi_modules[id].exports){
            continue;
        }
        for(var kk in pi_modules[id].exports){
            var c = pi_modules[id].exports[kk];
            if(!Struct.isPrototypeOf(c) || !c._$info || !c._$info.notes){
                continue;
            }
            let tab = c._$info.name;
            let ware = c._$info.notes.get("db");
            //let cfg = cfgs.get(ware);
            let path = `./dump/${DEFAULT_FILE_WARE}/${tab}`;
            if(ware === DEFAULT_FILE_WARE){ //判断是文件数据库
                try {
                    restore((<any>dbMgr).inner, ware, tab, path);
                } catch (error) {
                    console.log(error);
                }

            }
        }
    }
}

const writeCfg = (dbMgr: Mgr) => {
    let items = [];
    cfgMgr.map.forEach((tab, k) => {
        if (k.indexOf("#") < 0) { //如果不是一个主键表， 但类型中存在主键，则不处理该数据
            let first = tab.get(0);
            if(!first){
                return; //没有数据直接返回
            }

            let primary;
            first._$getSinfo().notes && (primary = first._$getSinfo().notes.get("primary"));
            if(primary){
                return;
            }
        }
        tab.forEach((value, key) => {
            let ware;
            value._$getSinfo().notes && (ware = value._$getSinfo().notes.get("db") || DEFAULT_WARE);
            if(!ware){
                throw "cfg type: " + value._$getSinfo().name + " not note 'db'"; //如果配置没有db注解， 将抛出异常
            }
            items.push({ware: ware, tab: value._$getSinfo().name, key: key, value: value});
        });
    });
    // let t1 =  new Date().getTime();
    if (items.length > 0) {
        dbMgr.write((tr) => {
            tr.modify(items, 1000, false);
        }, 500000);
    }
}

initDb();